"""
Declares all NetSuite types which are available through attribute lookup `ns.<type>`
of a :class:`~netsuitesdk.client.NetSuiteClient` instance `ns`.
"""

COMPLEX_TYPES = {
    'ns0': [
        'BaseRef',
        'GetAllRecord',
        'GetAllResult',
        'Passport',
        'RecordList',
        'RecordRef',
        'ListOrRecordRef',
        'SearchColumnStringField',
        'SearchColumnSelectField',
        'SearchResult',
        'SearchEnumMultiSelectField',
        'SearchStringField',
        'SearchMultiSelectField',
        'SearchDateField',
        'SearchLongField',
        'Status',
        'StatusDetail',
        'TokenPassport',
        'TokenPassportSignature',
        'WsRole',
        'DateCustomFieldRef',
        'CustomFieldList',
        'DoubleCustomFieldRef',
        'StringCustomFieldRef',
        'BooleanCustomFieldRef',
        'CustomRecordRef',
        'SelectCustomFieldRef'
    ],

    # ns4: https://webservices.netsuite.com/xsd/platform/v2017_2_0/messages.xsd
    'ns4': [
        'ApplicationInfo',
        'GetAllRequest',
        'GetRequest',
        'GetResponse',
        'GetAllResponse',
        'PartnerInfo',
        'ReadResponse',
        'SearchPreferences',
        'SearchResponse'
    ],

    # https://webservices.netsuite.com/xsd/platform/v2017_2_0/common.xsd
    'ns5': [
        'AccountSearchBasic',
        'Address',
        'CustomerSearchBasic',
        'JobSearchBasic',
        'LocationSearchBasic',
        'TransactionSearchBasic',
        'InboundShipmentSearchBasic',
        'VendorSearchBasic',
        'SubsidiarySearchBasic',
        'EmployeeSearchBasic',
        'FolderSearchBasic',
        'FileSearchBasic',
        'CustomRecordSearchBasic',
        'CustomListSearchBasic',
        'TermSearchBasic',
        'DepartmentSearchBasic',
        'ItemSearchBasic', 'ItemSearchRowBasic',
        'ClassificationSearchBasic',
        'CurrencyRateSearchBasic',
        'ConsolidatedExchangeRateSearchBasic',
        'BillingAccountSearchBasic'
    ],

    'ns6': [
      'ItemCostEstimateType',
    ],

    # urn:relationships.lists.webservices.netsuite.com
    'ns13': [
        'BillingAccount',
        'CustomerAddressbook', 'CustomerAddressbookList',
        'Customer', 'CustomerSearch', 'CustomerTaxRegistrationList', 'CustomerTaxRegistration',
        'Vendor', 'VendorSearch',
        'Job', 'JobSearch',
        'VendorAddressbook', 'VendorAddressbookList', 'BillingAccountSearch'
    ],

    # urn:accounting_2017_2.lists.webservices.netsuite.com
    # https://webservices.netsuite.com/xsd/lists/v2017_2_0/accounting.xsd
    'ns17': [
        'Account', 'AccountSearch',
        'ExpenseCategory', 'ExpenseCategorySearch',
        'AccountingPeriod',
        'Classification', 'ClassificationSearch',
        'Department', 'DepartmentSearch',
        'Location', 'LocationSearch',
        'Subsidiary', 'SubsidiarySearch',
        'VendorCategory', 'VendorCategorySearch',
        'Term', 'TermSearch',
        'SalesTaxItem', 'SalesTaxItemSearch',
        'TaxGroup', 'TaxGroupSearch',
        'DepartmentSearch',
        'ItemSearch', 'ItemSearchAdvanced', 'ItemSearchRow',
        'ClassificationSearch',
        'PriceLevelSearch'
    ],

    'ns19': [
        'Invoice',
        'InvoiceItem',
        'InvoiceItemList',
        'TransactionSearch',
        'TransactionSearchAdvanced',
        'Usage',
    ],

    # urn:purchases_2017_2.transactions.webservices.netsuite.com
    # https://webservices.netsuite.com/xsd/transactions/v2017_2_0/purchases.xsd
    'ns21': [
        'VendorBill',
        'VendorBillExpense',
        'VendorBillExpenseList',
        'VendorBillItem',
        'VendorBillItemList',
        'VendorCredit',
        'VendorCreditApply',
        'VendorCreditApplyList',
        'VendorCreditExpense',
        'VendorCreditExpenseList',
        'VendorCreditItem',
        'VendorCreditItemList',
        'VendorPayment',
        'VendorPaymentApplyList',
        'VendorPaymentCredit',
        'VendorPaymentCreditList',
        'InboundShipmentSearch',
        'VendorPaymentApply'
    ],

    'ns23': [
        'CreditMemo',
        'CreditMemoApply',
        'CreditMemoApplyList',
        'CreditMemoItem',
        'CreditMemoItemList',
    ],

    # urn:general_2019_2.transactions.webservices.netsuite.com
    # https://webservices.netsuite.com/xsd/transactions/v2019_2_0/general.xsd
    'ns31': [
        'JournalEntry',
        'JournalEntryLine',
        'JournalEntryLineList',
        'AdvInterCompanyJournalEntry',
        'AdvInterCompanyJournalEntryLine',
        'AdvInterCompanyJournalEntryLineList'
    ],

    'ns32': [
        'CustomRecord',
        'CustomRecordCustomField',
        'CustomRecordSearch',
        'CustomListSearch',
        'CustomRecordType'
    ],

    # https://webservices.netsuite.com/xsd/lists/v2019_2_0/employees.xsd
    'ns34': [
        'EmployeeSearch',
        'Employee'
    ],

    # urn:employees_2019_2.transactions.webservices.netsuite.com
    # https://webservices.netsuite.com/xsd/transactions/v2019_2_0/employees.xsd
    'ns38': [
        'ExpenseReport',
        'ExpenseReportExpense',
        'ExpenseReportExpenseList',
    ],
    'ns11': [
        'FolderSearch',
        'Folder',
        'File',
        'FileSearch'
    ],
}

SIMPLE_TYPES = {
    # ns1: view-source:https://webservices.netsuite.com/xsd/platform/v2017_2_0/coreTypes.xsd
    'ns1': [
        'RecordType',
        'GetAllRecordType',
        'SearchRecordType',
        'SearchEnumMultiSelectFieldOperator',
        'SearchStringFieldOperator',
        'SearchDateFieldOperator',
        'SearchLongFieldOperator',
    ],
}



from .base import ApiBase
import logging

logger = logging.getLogger(__name__)

class Departments(ApiBase):
    def __init__(self, ns_client):
        ApiBase.__init__(self, ns_client=ns_client, type_name='Department')



import zeep
import logging
from collections import OrderedDict

from netsuitesdk.internal.client import NetSuiteClient
from netsuitesdk.internal.utils import PaginatedSearch
from typing import List

logger = logging.getLogger(__name__)


# TODO: introduce arg and return types
class ApiBase:
    def __init__(self, ns_client: NetSuiteClient, type_name):
        self.ns_client = ns_client
        self.type_name = type_name

    def search(self, attribute, value, operator):
        """
        Search Record
        :param attribute: name of the field, eg. entityId
        :param value: value of the field, eg. Amazon
        :param operator: search matching operator, eg., 'contains', 'is', 'anyOf'
        :return:
        """
        records = self.ns_client.basic_stringfield_search(
            type_name=self.type_name,
            attribute=attribute,
            value=value,
            operator=operator
        )

        return records

    def get_all(self):
        generated_records = self.get_all_generator()
        all_records = []
        for records in generated_records:
            all_records.extend(records)
        return all_records

    def count(self):
        ps = PaginatedSearch(client=self.ns_client, type_name=self.type_name, pageSize=10, perform_search=True)
        return ps.total_records

    def get_all_generator(self, page_size=20):
        """
        Returns a generator which is more efficient memory-wise
        """
        return self.create_paginated_search(page_size=page_size)

    def get(self, internalId=None, externalId=None) -> OrderedDict:
        return self._get(internalId=internalId, externalId=externalId)

    def get_ref(self, internalId=None, externalId=None) -> OrderedDict:
        return self._serialize(self.ns_client.RecordRef(type=self.type_name.lower(),
                                                        internalId=internalId, externalId=externalId))

    def post(self, data) -> OrderedDict:
        raise NotImplementedError('post method not implemented')

    def delete(self, internalId=None, externalId=None) -> OrderedDict:
        record = self.ns_client.delete(recordType=self.type_name, internalId=internalId, externalId=externalId)
        return record

    def _serialize(self, record) -> OrderedDict:
        """
        record: single record
        Returns a dict
        """
        return zeep.helpers.serialize_object(record)

    def _serialize_array(self, records) -> List[OrderedDict]:
        """
        records: a list of records
        Returns an array of dicts
        """
        return zeep.helpers.serialize_object(records)

    @staticmethod
    def _paginated_search_to_generator(paginated_search):
        if paginated_search.num_records == 0:
            return

        num_pages = paginated_search.total_pages
        logger.debug('total pages = %d, records in page = %d', paginated_search.total_pages, paginated_search.num_records)
        logger.debug(f'current page index {paginated_search.page_index}')
        logger.debug('going to page %d', 0)

        records = []

        for p in range(1, num_pages + 1):
            logger.debug('going to page %d', p)
            paginated_search.goto_page(p)
            logger.debug(f'current page index {paginated_search.page_index}')
            records.extend(paginated_search.records)

        return records

    @staticmethod
    def _paginated_search_generator(paginated_search: PaginatedSearch):
        if paginated_search.num_records == 0:
            return

        num_pages = paginated_search.total_pages
        logger.debug('total pages = %d, records in page = %d', paginated_search.total_pages, paginated_search.num_records)
        logger.debug(f'current page index {paginated_search.page_index}')
        logger.debug('going to page %d', 0)

        for p in range(1, num_pages + 1):
            logger.debug('going to page %d', p)
            paginated_search.goto_page(p)
            logger.debug(f'current page index {paginated_search.page_index}')
            yield paginated_search.records

    def create_paginated_search(self, page_size):
        ps = PaginatedSearch(client=self.ns_client, type_name=self.type_name, pageSize=page_size)
        return self._paginated_search_generator(paginated_search=ps)

    def _search_all_generator(self, page_size):
        ps = PaginatedSearch(client=self.ns_client, type_name=self.type_name, pageSize=page_size)
        return self._paginated_search_to_generator(paginated_search=ps)

    def _get_all(self) -> List[OrderedDict]:
        records = self.ns_client.getAll(recordType=self.type_name)
        return self._serialize_array(records)

    def _get_all_generator(self):
        res = self._get_all()
        for r in res:
            yield r

    def _get(self, internalId=None, externalId=None) -> OrderedDict:
        record = self.ns_client.get(recordType=self.type_name, internalId=internalId, externalId=externalId)
        return record

    def build_simple_fields(self, fields, source, target):
        for field in fields:
            if field in source:
                target[field] = source[field]

    def build_record_ref_fields(self, fields, source, target):
        for field in fields:
            if field in source and source[field]:
                target[field] = self.ns_client.RecordRef(**(source[field]))



from collections import OrderedDict

from .base import ApiBase
import logging

logger = logging.getLogger(__name__)


class CustomRecords(ApiBase):
    SIMPLE_FIELDS = [
        'allowAttachments',
        'allowInlineEditing',
        'allowNumberingOverride',
        'allowQuickSearch',
        'altName',
        'autoName',
        'created',
        'customFieldList',
        'customRecordId',
        'description',
        'disclaimer',
        'enablEmailMerge',
        'enableNumbering',
        'includeName',
        'internalId',
        'isAvailableOffline',
        'isInactive',
        'isNumberingUpdateable',
        'isOrdered',
        'lastModified',
        'name',
        'numberingCurrentNumber',
        'numberingInit',
        'numberingMinDigits',
        'numberingPrefix',
        'numberingSuffix',
        'recordName',
        'scriptId',
        'showCreationDate',
        'showCreationDateOnList',
        'showId',
        'showLastModified',
        'showLastModifiedOnList',
        'showNotes',
        'showOwner',
        'showOwnerAllowChange',
        'showOwnerOnList',
        'translationsList',
        'usePermissions',
        'nullFieldList',
    ]

    RECORD_REF_FIELDS = [
        'customForm',
        'owner',
        'parent',
        'recType',
    ]

    def __init__(self, ns_client):
        ApiBase.__init__(self, ns_client=ns_client, type_name='CustomRecord')

    def post(self, data) -> OrderedDict:
        assert data['externalId'], 'missing external id'
        record = self.ns_client.CustomRecord(externalId=data['externalId'])

        self.build_simple_fields(self.SIMPLE_FIELDS, data, record)

        self.build_record_ref_fields(self.RECORD_REF_FIELDS, data, record)

        logger.debug('able to create custom record = %s', record)
        res = self.ns_client.upsert(record)
        return self._serialize(res)



from .journal_entries import JournalEntries
import logging

logger = logging.getLogger(__name__)


class AdvInterCompanyJournalEntries(JournalEntries):
    SIMPLE_FIELDS = [
        'memo',
        'tranDate',
        'tranId'
    ]

    RECORD_REF_FIELDS = [
        'class',
        'currency',
        'department',
        'location',
        'subsidiary',
        'toSubsidiary'
    ]

    TYPE_NAME = 'advInterCompanyJournalEntry'



class PaginatedSearch:

    default_page_size = 20

    def __init__(self, client, type_name, search_record=None, basic_search=None, pageSize=None, perform_search=True):
        """
        PaginatedSearch is a utility class that can be used to perform
        a search.

        :param NetSuiteClient client: the NetSuite client proxy object
        :param str type_name: the type name to perform the search on
        :param <record>Search search_record: object containing the search filters
        :param <record>SearchBasic basic_search: object containing basic search filters
        :param int pageSize: the number of records per result page
        :param bool perform_search: if True, will perform a first search after initializing basic data
        """

        self._ns = client
        self._result = None
        self._type_name = type_name
        self.search_record = search_record or self._ns.search_factory(type_name=self._type_name)
        self.basic_search = basic_search
        if self.basic_search is not None:
            self.search_record.basic = self.basic_search
        if perform_search:
            self.search()

    @property
    def total_records(self):
        return self._result.totalRecords

    @property
    def page_size(self):
        return self._result.pageSize

    @property
    def total_pages(self):
        return self._result.totalPages

    @property
    def page_index(self):
        return self._result.pageIndex

    @property
    def records(self):
        return self._result.records

    @property
    def num_records(self):
        if self.records:
            return len(self.records)
        return 0

    def search(self):
        """ Call the netsuite operation `search` """
        self._result = self._ns.search(searchRecord=self.search_record)

    def goto_page(self, page_index):
        """ After a search was performed, this method utilizes the NetSuite
        operation `searchMoreWithId` to retrieve more results """

        if self._result is None:
            return
        if page_index > self.total_pages or page_index < 1:
            return
        self._result = self._ns.searchMoreWithId(searchId=self._result.searchId,
                                                 pageIndex=page_index)
